defmodule Clickhouse do
  @moduledoc """
  A Clickhouse client.

  This currently represents an early work in progress.
  """

  @type client :: atom()
  @type statement :: binary()
  @type params :: list()
  @type interface :: module()
  @type format :: module()
  @type option ::
          {:client, client()}
          | {:interface, interface()}
          | {:interface_opts, keyword()}
  @type options :: [option()]

  use Supervisor

  alias Clickhouse.{Config, Query}

  @formats [
    {Clickhouse.Format.TSV, ["TabSeparated", "TSV"]},
    {Clickhouse.Format.TSVWithNames, ["TabSeparatedWithNames", "TSVWithNames"]},
    {Clickhouse.Format.TSVWithNamesAndTypes,
     ["TabSeparatedWithNamesAndTypes", "TSVWithNamesAndTypes"]}
  ]

  @opts_schema %{
    client: [type: :atom, required: true, default: :default],
    interface: [type: :module, required: true, default: Clickhouse.Interface.HTTP],
    interface_opts: [type: :list, required: true, default: []],
    formats: [
      type: {:list, {:tuple, {:module, {:list, :binary}}}},
      required: true,
      default: @formats
    ]
  }

  ################################
  # Public API
  ################################

  @doc """
  Starts a Clickhouse client.

  ## Options

    * `:client` - The name of the client - defaults to `:default`.
    * `:interface` - The interface to use for queries - defaults to `Clickhouse.Interface.HTTP`.
    * `:interface_opts` - Additional keyword options to pass to the interface.
    * `:formats` - A list of formats available for encoding/decoding.
  """
  @spec start_link(options()) :: Supervisor.on_start()
  def start_link(opts \\ []) do
    with {:ok, opts} <- validate_opts(opts, @opts_schema) do
      Supervisor.start_link(__MODULE__, opts)
    end
  end

  @doc """
  Prepares and executes a query using a Clickhouse client.
  """
  @spec query(client(), statement(), params(), keyword()) ::
          {:ok, Clickhouse.Result.t()} | {:error, Clickhouse.Error.t()}
  def query(client, statement, params \\ [], opts \\ []) do
    with {:ok, query} <- prepare(client, statement, params, opts) do
      execute(query)
    end
  end

  @doc """
  Prepares a query for a Clickhouse client.
  """
  @spec prepare(client(), statement(), params(), keyword()) ::
          {:ok, Clickhouse.Query.t()} | {:error, Clickhouse.Error.t()}
  def prepare(client, statement, params \\ [], opts \\ []) do
    query = Query.new(client, statement, params, opts)
    do_encode(query)
  end

  @doc """
  Executes a query using a Clickhouse client.
  """
  @spec execute(Clickhouse.Query.t()) ::
          {:ok, Clickhouse.Result.t()} | {:error, Clichouse.Error.t()}
  def execute(query) do
    with {:ok, result} <- do_execute(query) do
      do_decode(result)
    end
  end

  ################################
  # Supervisor Callbacks
  ################################

  @impl Supervisor
  def init(opts) do
    init_config(opts)

    children = [
      build_interface(opts)
    ]

    Supervisor.init(children, strategy: :one_for_one)
  end

  ################################
  # Private API
  ################################

  defp validate_opts(opts, schema) do
    case KeywordValidator.validate(opts, schema) do
      {:ok, _} = result -> result
      {:error, reason} -> {:error, reason}
    end
  end

  defp init_config(opts) do
    client = Keyword.fetch!(opts, :client)
    Config.new(client, opts)
  end

  defp build_interface(opts) do
    client = Keyword.fetch!(opts, :client)
    interface = Keyword.fetch!(opts, :interface)
    interface_opts = Keyword.fetch!(opts, :interface_opts)
    interface_opts = Keyword.put(interface_opts, :client, client)

    {interface, interface_opts}
  end

  defp do_encode(%{format: nil} = query), do: {:ok, query}
  defp do_encode(query), do: query.format.encode(query)

  defp do_execute(%{interface: nil}), do: raise(ArgumentError, "No query interface")
  defp do_execute(query), do: query.interface.execute(query)

  defp do_decode(%{format: nil} = result), do: {:ok, result}
  defp do_decode(result), do: result.format.decode(result)
end
